Festival = LibStub("AceAddon-3.0"):NewAddon("Festival", "AceEvent-3.0", "AceConsole-3.0")

local function tcount(T)
  local count = 0
  for _ in pairs(T) do count = count + 1 end
  return count
end

local function pdebug(str, ...)
	-- if ... then
		-- print(format(str, ...))
	-- else
		-- print(str)
	-- end
end

local function ConvertItemLinkToID(link)
	return tonumber(string.match(link, "item:(%d+):"))
end
	
--[[ 	GetItemInfo is likely to be forced to fetch a value from the server, so
		I choose to wrap the function in a fancy callback mechanism.
--]]
local ItemScanner = CreateFrame("Frame")
local Blizzard_GetItemInfo = GetItemInfo
ItemScanner:RegisterEvent("GET_ITEM_INFO_RECEIVED")
ItemScanner.callbacks = {}
local function GetItemInfo(link, callback)
	if not Blizzard_GetItemInfo(link) then
		if not ItemScanner.callbacks[link] then
			ItemScanner.callbacks[link] = {}
		end
		table.insert(ItemScanner.callbacks[link], callback)
	else
		callback(Blizzard_GetItemInfo(link))
	end
end
ItemScanner:SetScript("OnEvent", function(self, event, ...)
	if event == "GET_ITEM_INFO_RECEIVED" then
		for link, linkCallbacks in pairs(self.callbacks) do
			if Blizzard_GetItemInfo(link) then
				for _, v in pairs(linkCallbacks) do
					v(Blizzard_GetItemInfo(link))
				end
				self.callbacks[link] = nil
			end
		end
	end
end)

local primaryItem = 101662
local items = {
	[101662] = {
		[74646] = 5,
		[74656] = 5,
		[74650] = 5,
		[74653] = 5,
		[74648] = 5,
		[75038] = 5,
		},
	[74646] = {
		[74840] = 25,
		[74834] = 5,
		[74857] = 5,
		[74661] = 1,
		},
	[74656] = {
		[74844] = 25,
		[74833] = 5,
		[74860] = 5,
		[74662] = 1,
		},
	[74650] = {
		[74842] = 25,
		[75014] = 5,
		[74859] = 5,
		[74661] = 1,
		},
	[74653] = {
		[74850] = 25,
		[74838] = 5,
		[74863] = 5,
		[74662] = 1,
		},
	[74648] = {
		[74843] = 25,
		[74837] = 5,
		[74861] = 5,
		[74662] = 1,
		},
	[75038] = {
		[74849] = 50,
		[74661] = 1,
		},
}

Festival.PADDING = 8
Festival.ROW_HEIGHT = 16
Festival.ROW_PADDING = 2
Festival.COLUMN_WIDTH = 75 -- this is bad and dumb but this gets set once in the following function
Festival.COLUMN_PADDING = 4
Festival.HEADER_HEIGHT = 3 -- in terms of rows
function Festival:CreateItemEntry(itemID, column, row, xOff, yOff)
	local frame = self.frame
	local buttonName = "FestivalButtonColumn"..column.."Row"..row

	local button = CreateFrame("Button", buttonName, frame);
	
	-- pdebug("Topleft (%s, %s), bottomright (%s,%s)",self.PADDING + (column*(self.COLUMN_WIDTH+self.COLUMN_PADDING)) + xOff,
					-- -self.PADDING - (row+self.HEADER_HEIGHT)*(self.ROW_HEIGHT+self.ROW_PADDING) - yOff, self.PADDING + (column*(self.COLUMN_WIDTH+self.COLUMN_PADDING)) + xOff + self.ROW_HEIGHT,
						-- -self.PADDING - (row+self.HEADER_HEIGHT)*(self.ROW_HEIGHT+self.ROW_PADDING) - yOff - self.ROW_HEIGHT)
	button:SetWidth(self.ROW_HEIGHT);
	button:SetHeight(self.ROW_HEIGHT);
	button:SetPoint("TOPLEFT", frame, "TOPLEFT",
					self.PADDING + (column*(self.COLUMN_WIDTH+self.COLUMN_PADDING)) + xOff,
					-self.PADDING - (row+self.HEADER_HEIGHT)*(self.ROW_HEIGHT+self.ROW_PADDING) - yOff);
	button:SetHighlightTexture([[Interface\Buttons\ButtonHilight-Round]]);
	
	
	GetItemInfo(itemID, function(...)
		-- for k,v in pairs({...}) do
			-- print(tostring(k)..": "..tostring(v))
		-- end
		local _, link, _, _, _, _, _, _, _, texture = ...
		-- -- pdebug("Got texture "..tostring(texture))
		button:SetNormalTexture(texture);
		button:SetPushedTexture(texture);
		button.link = link
	end)
	
	button:SetScript("OnEnter", function(self)
		if self.link then
			-- -- pdebug("Setting tooltip to %s", self.link)
			GameTooltip:SetOwner(self, "ANCHOR_LEFT")
			GameTooltip:SetHyperlink(self.link)
			GameTooltip:Show()
		end
	end)
	button:SetScript("OnLeave", function(self)
		GameTooltip:Hide()
	end)
	
	local guildCount = frame:CreateFontString(buttonName.."GuildCount", "ARTWORK")
	guildCount:SetFontObject("GameFontHighlightSmall")
	guildCount:SetPoint("LEFT", button, "RIGHT", self.COLUMN_PADDING, 0)
	guildCount:SetJustifyH("LEFT")
	
	guildCount:SetText("Guild")
	local stringWidth = guildCount:GetStringWidth()
	self.stringWidth = stringWidth
	self.COLUMN_WIDTH = stringWidth*3 + self.ROW_HEIGHT + self.COLUMN_PADDING*4
	
	guildCount:SetWidth(stringWidth)
	local selfCount = frame:CreateFontString(buttonName.."SelfCount", "ARTWORK")
	selfCount:SetFontObject("GameFontHighlightSmall")
	selfCount:SetPoint("LEFT", guildCount, "RIGHT", self.COLUMN_PADDING, 0)
	selfCount:SetText("444")
	selfCount:SetJustifyH("LEFT")
	selfCount:SetWidth(stringWidth)
	
	local totalCount = frame:CreateFontString(buttonName.."TotalCount", "ARTWORK")
	totalCount:SetFontObject("GameFontHighlightSmall")
	totalCount:SetPoint("LEFT", selfCount, "RIGHT", self.COLUMN_PADDING, 0)
	totalCount:SetText("444")
	totalCount:SetWidth(stringWidth)
	totalCount:SetJustifyH("LEFT")
	
	if row % 2 == 0 then
		-- pdebug("Making "..buttonName.." stripe!")
		-- local buttonBackground = CreateFrame("Frame")
		--== Create zebra striping
		local stripe = frame:CreateTexture(buttonName.."Stripe")
		stripe:SetTexture(1, 1, 1, 0.07)
		stripe:SetPoint("TOPLEFT", frame, "TOPLEFT", self.PADDING + (column*(self.COLUMN_WIDTH+self.COLUMN_PADDING)) + xOff,
						-self.PADDING - (row+self.HEADER_HEIGHT)*(self.ROW_HEIGHT+self.ROW_PADDING) + self.ROW_PADDING/2 - yOff)
		stripe:SetPoint("BOTTOMRIGHT", frame, "TOPLEFT", self.PADDING + ((column+1)*(self.COLUMN_WIDTH+self.COLUMN_PADDING)) + xOff - self.COLUMN_PADDING/2,
						-self.PADDING - (row+self.HEADER_HEIGHT)*(self.ROW_HEIGHT+self.ROW_PADDING) - self.ROW_PADDING/2 - yOff - self.ROW_HEIGHT)
						
	end
	
	if not self.displayEntries[itemID] then
		self.displayEntries[itemID] = {
			buttons = {},
			guildCounts = {},
			selfCounts = {},
			totalCounts = {},
		}
	end
	
	tinsert(self.displayEntries[itemID].buttons, button)
	tinsert(self.displayEntries[itemID].guildCounts, guildCount)
	tinsert(self.displayEntries[itemID].selfCounts, selfCount)
	tinsert(self.displayEntries[itemID].totalCounts, totalCount)
end

function Festival:CreateRecipeEntry(itemID, reagents, column, row, xOff, yOff)
	local frame = self.frame
	xOff, yOff = (xOff or 0), (yOff or 0)
	column = column or 0
	row = row or 0
	self:CreateItemEntry(itemID, column, row, xOff, yOff)
	
	column = column + 1
	local sortedKeys = {}
	for k,v in pairs(reagents) do table.insert(sortedKeys, k) end
	table.sort(sortedKeys)
	for _, innerItem in ipairs(sortedKeys) do
	-- for innerItem, _ in pairs(reagents) do
		-- Is there another level?
		if items[innerItem] then
			self:CreateRecipeEntry(innerItem, items[innerItem], column, row, xOff, yOff)
			row = row + tcount(items[innerItem])
			yOff = yOff + self.ROW_HEIGHT/2
			self.spacerRows = self.spacerRows + 1
		-- Otherwise, it's a final reagent
		else
			self:CreateItemEntry(innerItem, column, row, xOff, yOff)
			row = row + 1
		end
	end
	if row > self.rows then
		self.rows = row
	end
	if column >= self.columns then
		self.columns = column
	end
end

function Festival:CreateHeaders()
	local headers = {"Guild", "Self", "Total"}
	for i = 0, self.columns do
		for j = 1, 3 do
			local columnHeader = self.frame:CreateFontString("FestivalButtonColumn"..i.."Header"..j, "ARTWORK")
			columnHeader:SetFontObject("GameFontNormalSmall")
			local text = headers[j]
			-- columnHeader:SetPoint("BOTTOMLEFT", "FestivalButtonColumn"..i.."Row0", "TOPRIGHT",
								-- (self.COLUMN_PADDING + self.stringWidth)*j,
								-- (self.ROW_HEIGHT+self.ROW_PADDING))
			columnHeader:SetPoint("BOTTOMLEFT", "FestivalButtonColumn"..i.."Row0"..text.."Count", "TOPLEFT", 0, 2*self.ROW_PADDING)
			columnHeader:SetText(headers[j])
		end
	end
end

function Festival:CreateInstructions()
	local general = self.frame:CreateFontString("FestivalNote", "ARTWORK")
	general:SetFontObject("GameFontNormalLarge")
	general:SetPoint("TOP", self.frame, "TOP", 0, - self.PADDING) --self.PADDING, -self.PADDING) --self.COLUMN_PADDING, self.ROW_PADDING)
	self.frame.general = general
	general:SetText("Festival")
	-- general:SetWidth(self.frame:GetWidth())
	general:SetJustifyH("CENTER")
	
	local lastSeen = self.frame:CreateFontString("FestivalLastSeen", "ARTWORK")
	lastSeen:SetFontObject("GameFontHighlightSmall")
	lastSeen:SetPoint("TOPLEFT", self.frame, "TOPLEFT", self.PADDING, -(self.PADDING + self.ROW_PADDING + self.ROW_HEIGHT)) --self.COLUMN_PADDING, self.ROW_PADDING)
	self.frame.lastSeen = lastSeen
	
	local lastSeenWarning = self.frame:CreateFontString("FestivalLastSeenWarning", "ARTWORK")
	lastSeenWarning:SetFontObject("GameFontHighlightSmall")
	lastSeenWarning:SetPoint("TOPLEFT", self.frame, "TOPLEFT", self.PADDING, -(self.PADDING + 2*self.ROW_HEIGHT)+self.ROW_PADDING) --self.COLUMN_PADDING, self.ROW_PADDING)
	self.frame.lastSeenWarning = lastSeenWarning
	
	local scanButton = CreateFrame("Button", "FestivalScanButton", self.frame);
	
	scanButton:SetWidth(self.ROW_HEIGHT*2);
	scanButton:SetHeight(self.ROW_HEIGHT*2);
	scanButton:SetPoint("TOPRIGHT", self.frame, "TOPRIGHT", -self.PADDING, -self.PADDING);
	scanButton:SetNormalTexture([[Interface\Buttons\UI-SquareButton-Up]]);
	scanButton:SetPushedTexture([[Interface\Buttons\UI-SquareButton-Down]]);
	scanButton:SetHighlightTexture([[Interface\Buttons\UI-Common-MouseHilight]]);
	local tex = scanButton:CreateTexture("FestivalScanButtonIcon", "OVERLAY")
	tex:SetSize(Festival.ROW_HEIGHT, Festival.ROW_HEIGHT)
	tex:SetPoint("CENTER")
	tex:SetTexture([[Interface\Buttons\UI-RefreshButton]])
	scanButton.tex = tex
	
	scanButton:SetScript("OnClick", function()
		PlaySound("igMainMenuOptionCheckBoxOn");
		Festival:ScanAllTabs()
	end)
	scanButton:SetScript("OnMouseDown", function(self)
		self.tex:SetPoint("CENTER", self, "CENTER", -1, -2)
	end)
	scanButton:SetScript("OnMouseUp", function(self)
		self.tex:SetPoint("CENTER", self, "CENTER")
	end)
	scanButton:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		GameTooltip:SetText("Click for full guild bank scan")
		GameTooltip:Show()
	end)
	scanButton:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
	self.frame.scanButton = scanButton
end

-- The frame and x/y location to SetPoint to if a guild bank replacement is loaded.
-- For LOD addons, the fourth element is the function that creates the guild bank replacement, if necessary.
-- Fifth and later elements are expected values for that function
local guildBankReplacements = {
	["Bagnon_GuildBank"] = {
		frame = "BagnonFrameguildbank",
		x = 0,
		y = 0,
		creationFunction = "Bagnon.Frame.New",
		creationArgs = "guildbank",
		tabName = "BagnonGuildTab%d",
		popButtonX = -4,
		popButtonY = -30,
		popButtonSelfPoint = "TOPRIGHT",
	},
	["Blizzard_GuildBankUI"] = {
		frame = "GuildBankFrame",
		x = 44,
		y = 0,
		tabName = "GuildBankTab%dButton",
		popButtonX = 2,
		popButtonY = -6,
		popButtonSelfPoint = "LEFT",
	},
}

function Festival:ShowPopButton()
	if not self.popButton then
		self:CreatePopButton()
	end
			
	self.popButton:ClearAllPoints()
	self.popButton:SetPoint(self.popButtonSelfPoint, _G[self.attachFrame], "TOPRIGHT", self.popButtonX, self.popButtonY)
	-- pdebug("Showing pop")
	if self.db.global.autoPop then
		-- pdebug("hiding pop lock")
		self.popButton.locked:Hide()
	else	
		-- pdebug("showing pop lock")
		self.popButton.locked:Show()
	end
	-- Festival.popButton:SetPoint(Festival.popButtonSelfPoint, _G[Festival.attachFrame], "TOPRIGHT", Festival.popButtonX, Festival.popButtonY)
end

function Festival:LoadAttachment(settings)
	for addon, settings in pairs(guildBankReplacements) do
		if IsAddOnLoaded(addon) then
			if not (self.settings or (self.settings and self.settings.attachFrame ~= settings.frame)) then
				self.attachFrame = settings.frame
				self.attachX = settings.x
				self.attachY = settings.y
				self.tabName = settings.tabName
				self.popButtonX = settings.popButtonX
				self.popButtonY = settings.popButtonY
				self.popButtonSelfPoint = settings.popButtonSelfPoint
				self.settings = settings
			end
		end
	end
end

function Festival:ApplyAttachment()
	self.frame:ClearAllPoints()
	self.frame:SetPoint("TOPLEFT", _G[self.attachFrame], "TOPRIGHT", self.attachX, self.attachY)
	-- pdebug("%s, %s, %s, %d, %d", self.popButtonSelfPoint, _G[self.attachFrame], "TOPRIGHT", self.popButtonX, self.popButtonY)
end

-- Some assholes like Bagnon use Lua to create their frames and that means waiting quite a while...
-- If the addon has a creation function, hook that, otherwise attach immediately
function Festival:AsyncAddOnPosthook(f, a)
	local settings = self.settings
	if not _G[settings.frame] and settings.creationFunction then
		local tab = _G
		gsub(settings.creationFunction, "([^.]+)%.", function(s) tab = tab[s] end)
		local func = string.match(settings.creationFunction, "%.([^.]+)$")
		hooksecurefunc(tab, func, function(...)
			local _, id = ...
			if settings.creationArgs == id then
				f(a)
			end
		end)
	else
		f(a)
	end
end

function Festival:AnchorFrame()
	self:AsyncAddOnPosthook(self.ApplyAttachment, self)
	self.frameAnchored = true
	self:ResizeFrame()
end

function Festival:ResizeFrame()
	self.frame:SetScale(UIParent:GetScale())
	-- The padding is *2/2 for a mnemonic. The real padding is half that, but inflated for visual.
	self.frame:SetHeight((self.ROW_HEIGHT+self.ROW_PADDING) * (self.rows+self.HEADER_HEIGHT) + self.PADDING*2/2 +self.spacerRows*(self.ROW_HEIGHT/2))
	self.frame:SetWidth((self.COLUMN_WIDTH+self.COLUMN_PADDING) * (self.columns+1) + self.COLUMN_PADDING + self.PADDING*2/2)
end

function Festival:ToggleAutoPop()
	-- pdebug("Toggling autopop")
	if self.db.global.autoPop then
		self.db.global.autoPop = false
		self.popButton.locked:Show()
	else
		self.db.global.autoPop = true
		self.popButton.locked:Hide()
	end
end

function Festival:CreatePopButton()
	local popButton = CreateFrame("Button", "FestivalPopButton", _G[self.attachFrame]);
	-- pdebug("help")
	popButton:SetSize(36,36);
	-- popButton:SetPoint("TOPRIGHT", self.frame, "TOPRIGHT", -self.PADDING, -self.PADDING);
	popButton:SetNormalTexture([[Interface\Buttons\UI-SquareButton-Up]]);
	popButton:SetPushedTexture([[Interface\Buttons\UI-SquareButton-Down]]);
	popButton:SetHighlightTexture([[Interface\Buttons\ButtonHilight-Round]]);
	popButton:RegisterForClicks("AnyUp")
	local tex = popButton:CreateTexture("FestivalPopButtonIcon", "OVERLAY")
	tex:SetSize(16, 16)
	tex:SetPoint("CENTER")
	tex:SetTexture([[Interface\ICONS\Achievement_WorldEvent_Thanksgiving]])
	popButton.tex = tex
	local _, n = tex:GetDrawLayer()
	
	local locked = popButton:CreateTexture("FestivalPopButtonIconLocked", "OVERLAY")
	locked:SetSize(24, 24)
	locked:SetPoint("CENTER")
	locked:SetTexture([[Interface\PaperDollInfoFrame\UI-GearManager-LeaveItem-Transparent]])
	locked:SetDrawLayer("OVERLAY", n+1)
	popButton.locked = locked
	
	popButton:SetScript("OnMouseDown", function(self)
		self.tex:SetPoint("CENTER", self, "CENTER", -1, -2)
		self.locked:SetPoint("CENTER", self, "CENTER", -1, -2)
	end)
	popButton:SetScript("OnMouseUp", function(self)
		self.tex:SetPoint("CENTER", self, "CENTER")
		self.locked:SetPoint("CENTER", self, "CENTER")
	end)
	popButton:SetScript("OnClick", function(self, mouseButton)
		PlaySound("igMainMenuOptionCheckBoxOn");
		if mouseButton == "RightButton" then
			Festival:ToggleAutoPop()
		elseif mouseButton == "LeftButton" then
			Festival:ToggleFrame()
		end
	end)
	popButton:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		GameTooltip:SetText("Toggle Festival")
		GameTooltip:AddLine("Right click to disable auto-open")
		GameTooltip:Show()
	end)
	popButton:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
	
	self.popButton = popButton
end

function Festival:OnInitialize()
	local dbdefaults = {
		profile = {
			guildLastSeen = {},
			guildContents = {},
			bagContents = {},
			bankContents = {},
		},
		global = {
			autoPop = true,
		},
	}
	self.db = LibStub("AceDB-3.0"):New("FestivalDB", dbdefaults)
	self:RegisterChatCommand("festival", "SlashCmd")
	self:RegisterChatCommand("fest", "SlashCmd")
end

function Festival:OnEnable()
	self:RegisterEvent("GUILDBANKFRAME_OPENED")
	self:RegisterEvent("GUILDBANKFRAME_CLOSED")
	self:RegisterEvent("GUILDBANKBAGSLOTS_CHANGED")
	self:RegisterEvent("GUILD_ROSTER_UPDATE")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("BAG_UPDATE")
	
	local FestivalFrame = CreateFrame("Frame", "FestivalFrame")
	FestivalFrame:SetWidth(200);
	FestivalFrame:SetHeight(364);
	FestivalFrame:EnableMouse(true);
	FestivalFrame:SetFrameStrata("HIGH")
	FestivalFrame:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 4, right = 4, top = 4, bottom = 4 }
	})
	FestivalFrame:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
	-- FestivalFrame:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
	FestivalFrame:SetBackdropColor(0, 0, 0);
	FestivalFrame:SetClampedToScreen(true)
	FestivalFrame:Hide()

	FestivalFrame:SetScript("OnMouseUp", function(self, button)
		if button == "RightButton" then
			Festival:CloseFrame()
		end
	end)	
	
	self.frame = FestivalFrame
	self.originalTab = 0
	self.currentSaved = {}
	self.isSaving = 0
	self.displayEntries = {}
	self.columns = 0
	self.rows = 0
	self.spacerRows = 0
	self:CreateRecipeEntry(primaryItem, items[primaryItem])
	self:CreateHeaders()
	self:CreateInstructions()
	self:SetGuildBankOpen(false)
	
	---- pdebug("%d rows, %d columns", self.rows, self.columns)
	
end

-- The player's guild info isn't available until a GRU sometime soon after login.
-- However, if another guild member logs in immediately after the player,
-- There'll be a GRU at which there's still no guild info available.
-- Strangely, IsInGuild is correct from the beginning.
function Festival:GUILD_ROSTER_UPDATE()
	if not self.doOnce then
		 -- pdebug(GetGuildInfo("player"))
		if IsInGuild() then
			if GetGuildInfo("player") then
				self.db:SetProfile(GetGuildInfo("player") .. " ! " .. GetRealmName())
				self.doOnce = true
			else
				return
			end
		else
			self.db:SetProfile("!GUILDLESS!" .. " ! " .. GetRealmName())
		end
	end
end

function Festival:PLAYER_ENTERING_WORLD()
	self:LoadAttachment()
	
	-- Because fuck you Bagnon
	if not MAX_GUILDBANK_SLOTS_PER_TAB then
		MAX_GUILDBANK_SLOTS_PER_TAB = 98
	end
end

function Festival:IsFrameAnchored()
	return self.frameAnchored
end

function Festival:AttachToGuildBankFrame()
	self.frame:ClearAllPoints()
	self:AnchorFrame()
	self.frame:Show()
end

function Festival:SetGuildBankOpen(bool)
	self.guildBankOpen = bool
	if self.guildBankOpen then
		self.frame.scanButton:Show()
	else
		self.frame.scanButton:Hide()
	end
end

function Festival:GUILDBANKFRAME_OPENED()
	pdebug("GBF open")
	self:LoadAttachment()
	self:SetGuildBankOpen(true)
	self:AsyncAddOnPosthook(function()	
		self:ShowPopButton()
		if self.db.global.autoPop then
			if not self.frame:IsShown() then
				self:AttachToGuildBankFrame()
			end
			self:UpdateDisplay()
		end
	end)
end

function Festival:SlashCmd()
	self:ToggleFrame()
end

function Festival:CloseFrame()
	self.frameAnchored = false
	self.frame:Hide()
end

function Festival:OpenFrame()
	local frame = self.frame
	self:ResizeFrame()
	self:AsyncAddOnPosthook(self.UpdateDisplay, self)
	
	if self:IsFrameAnchored() then
		frame:SetMovable(false)
		frame:RegisterForDrag(nil)
		frame:SetScript("OnDragStart", nil)
		frame:SetScript("OnDragStop", nil)
	else
		frame:SetMovable(true)
		frame:RegisterForDrag("LeftButton")
		frame:SetScript("OnDragStart", frame.StartMoving)
		frame:SetScript("OnDragStop", frame.StopMovingOrSizing)
		frame:SetPoint("CENTER", UIParent, "CENTER")
	end
	
	frame:Show()
end

function Festival:ToggleFrame()
	if self.frame:IsShown() then
		Festival:CloseFrame()
	else
		Festival:OpenFrame()
	end
end

function Festival:GUILDBANKFRAME_CLOSED()
	if self:IsFrameAnchored() then
		self:CloseFrame()
	end
	self:SetGuildBankOpen(false)
end

function Festival:GUILDBANKBAGSLOTS_CHANGED()
	if not self.frame:IsShown() then return end
	self:SaveCurrentGuildBankTab()
	if self.isSaving == 1 then			
		self:LoadNextGuildBankTab()
		SetCurrentGuildBankTab(self.nextTab)
		-- -- pdebug("Setting 'current' to %d", self.nextTab)
	elseif self.isSaving == 2 then
		SetCurrentGuildBankTab(self.originalTab)
		self:EnableGuildBankButtons()
		self.isSaving = 0
	end
end

function Festival:BAG_UPDATE()
	self:SaveBags()
end

function Festival:UpdateEntryValues(entry, linkID)
	local guildCount, selfCount, totalCount = 0, 0, 0
	for k,v in pairs(self.db.profile.guildContents) do
		if v[linkID] then
			guildCount = guildCount + v[linkID]
		end
	end
	for k,v in pairs(self.db.profile.bagContents) do
		if v[linkID] then
			selfCount = selfCount + v[linkID]
		end
	end
	totalCount = guildCount + selfCount
	
	entry.guildCountValue = guildCount
	entry.selfCountValue = selfCount
	entry.totalCountValue = totalCount
	
	return totalCount
end

function Festival:UpdateEntryText(entry, linkID, m)
	local f
	if entry.totalCountValue < 0.20*m then
		f = RED_FONT_COLOR_CODE
	elseif entry.totalCountValue < 0.8*m then
		f = YELLOW_FONT_COLOR_CODE
	else
		f = GREEN_FONT_COLOR_CODE
	end
	
	for i, gC in ipairs(entry.guildCounts) do
		gC:SetText(format("%s%s%s|r", f, entry.guildCountValue, ","))
	end
	for i, sC in ipairs(entry.selfCounts) do
		sC:SetText(format("%s%s%s|r", f, entry.selfCountValue, ","))
	end
	for i, tC in ipairs(entry.totalCounts) do
		tC:SetText(format("%s%s%s|r", f, entry.totalCountValue, ""))
	end
end

local function numberForRecipe(itemID)
	local t = 0
	for recipe, mats in pairs(items) do
		if mats[itemID] then
			t = t + mats[itemID]
		end
	end
	if t > 0 then
		return t
	else
		return -1
	end
end

local function getItemParent(itemID)
	for i, reagents in pairs(items) do
		if reagents[itemID] then
			return i
		end
	end
end

function Festival:MostMakeable(item, totals)
	-- It's an end-item, sub-component
	local leastReagent = math.huge
	local makeable = 0
	for reagentID, count in pairs(items[item]) do
		makeable = totals[reagentID]/count
		if makeable < leastReagent then
			leastReagent = makeable
		end
	end
	-- pdebug("Can only make %d of %d", leastReagent, item)
	return leastReagent
end

function Festival:TheoreticalMakeable(item, totals)
	local theoretical = 0
	local mats = items[item]
	
	if mats then
		-- It's a recipe, component of the noodle cart
		local makeable, current, sum = 0, 0, 0
		for reagentID, count in pairs(mats) do
			current = totals[reagentID]
			makeable = self:MostMakeable(reagentID, totals)
			sum = (current+makeable)/items[item][reagentID];
			-- pdebug("Have %d and can make %d of %d, equals %d carts", current, makeable, reagentID, sum)
			if sum > theoretical then
				theoretical =  sum
			end
		end
		
		-- pdebug("Can create %d of %d", sum, item)
		return theoretical
	-- else

	end
end

function Festival:UpdateDisplay()
	local totals = {}
	-- Update the values for each itemID
	for k, v in pairs(self.displayEntries) do
		totals[k] = self:UpdateEntryValues(v, k)
		-- Now that the display is set, normalize for multi-items like rice flour
		local reps = #v.totalCounts
		-- -- pdebug("TOTAL %d of %d, %d reps, setting to %d", totals[k], k, reps, math.floor(totals[k]/reps))
		totals[k] = math.floor(totals[k]/reps)
	end
	
	-- Colorize the display using the normalized item totals
	local theoretical = math.floor(self:TheoreticalMakeable(primaryItem, totals))
	
	 -- pdebug("Most makeable %d", theoretical)
	for k, v in pairs(self.displayEntries) do
		local needed = theoretical
		if items[k] then
			if k ~= primaryItem then
				needed = (theoretical*5 - totals[k])
			end
		else
			local parent = getItemParent(k)
			needed = ((theoretical*5-totals[parent])*numberForRecipe(k)) - totals[k]
			 -- pdebug("%d: ((%d*5 - %d)*%d - %d) = %d", k, theoretical, totals[parent], numberForRecipe(k), totals[k], needed)
		end
		self:UpdateEntryText(v, k, needed)
	end
	
	self:UpdateLastSeen()
end

function Festival:UpdateLastSeen()
	local t = time()
	local fontColor
	local oldestTab = math.huge
	for tab, lastSeen in pairs(self.db.profile.guildLastSeen) do
		if lastSeen < oldestTab then
			oldestTab = lastSeen
		end
	end
	-- If this is true, no gank scan yet
	if oldestTab == math.huge then return end
	local s = self
	local sa = self.attachFrame
	local sag = _G[sa]
	if sa and not sag:IsShown() then
		fontColor = RED_FONT_COLOR_CODE
		self.frame.lastSeenWarning:SetText("Guild bank data cannot be updated while away from the bank.")
	else
		if t - oldestTab > 60*60*24 then
			fontColor = RED_FONT_COLOR_CODE
			self.frame.lastSeenWarning:SetText("Guild bank data likely outdated - please rescan!")
		elseif t - oldestTab > 60*60 then
			fontColor = YELLOW_FONT_COLOR_CODE
			self.frame.lastSeenWarning:SetText("Guild bank data may be slightly dated.")
		else
			fontColor = GREEN_FONT_COLOR_CODE
			self.frame.lastSeenWarning:SetText("")
		end
	end
	-- pdebug(oldestTab)
	self.frame.lastSeen:SetText(format("Guild bank last scanned at %s%s|r", fontColor, date("%I:%M%p, %x", oldestTab)))
end

function Festival:SaveBags()
	wipe(self.db.profile.bagContents)
	
	local _, count, link, itemID

	for bag = 0, NUM_BAG_SLOTS do
		self.db.profile.bagContents[bag] = {}
		for slot = 1, GetContainerNumSlots(bag) do
			_, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)
			if link then
				itemID = ConvertItemLinkToID(link)
				-- Because fuck you, Christmas.
				if itemID then
					if self.db.profile.bagContents[bag][itemID] then
						self.db.profile.bagContents[bag][itemID] = self.db.profile.bagContents[bag][itemID] + count
					else
						self.db.profile.bagContents[bag][itemID] = count
					end
				end
			end
		end
	end
	
	if self.frame:IsShown() then
		self:AsyncAddOnPosthook(self.UpdateDisplay, self)
	end
end

function Festival:SaveBank()

end

function Festival:SaveCurrentGuildBankTab()
	local db = self.db.profile
	local _, link, linkID
	local currentTab = GetCurrentGuildBankTab()
	-- pdebug("Saving tab %d", currentTab)
	
	db.guildLastSeen[currentTab] = time()
	db.guildContents[currentTab] = {}
	for i = 1, MAX_GUILDBANK_SLOTS_PER_TAB do
		link = GetGuildBankItemLink(currentTab, i)
		if link then
			linkID = ConvertItemLinkToID(link)
			_, count = GetGuildBankItemInfo(currentTab, i)
			if db.guildContents[currentTab][linkID] then
				db.guildContents[currentTab][linkID] = db.guildContents[currentTab][linkID] + count
			else
				db.guildContents[currentTab][linkID] = count
			end
			-- if self.displayEntries[linkID] then
				-- self:UpdateEntry(linkID)
			-- end
		end
	end
	
	self:AsyncAddOnPosthook(self.UpdateDisplay, self)
	self.currentSaved[GetCurrentGuildBankTab()] = true
end

function Festival:LoadNextGuildBankTab()
	local numTabs = GetNumGuildBankTabs()
	for i = 1, numTabs do
		if not self.currentSaved[i] then
			-- -- pdebug("Found un-saved tab at %d", i)
			QueryGuildBankTab(i)
			self.nextTab = i
			return
		end
	end
	-- pdebug("Save process finished")
	self.isSaving = 2
	-- Re-query the tab to reset the position
	QueryGuildBankTab(self.originalTab)
end

function Festival:DisableGuildBankButtons()
	for i = 1, GetNumGuildBankTabs() do
		_G[format(self.tabName, i)]:Disable()
	end
end

function Festival:EnableGuildBankButtons()
	for i = 1, GetNumGuildBankTabs() do
		_G[format(self.tabName, i)]:Enable()
	end
end

function Festival:ScanAllTabs()
	self:SaveBags()
	
	self:DisableGuildBankButtons()
	self.originalTab = GetCurrentGuildBankTab()
	-- pdebug("Starting at tab %s", self.originalTab)
	self.currentSaved = {}
	self.isSaving = 1
	
	self:LoadNextGuildBankTab()
	-- SetCurrentGuildBankTab(self.originalTab)
end
